# This file is copied from Intel's documentation: https://intel.github.io/intel-extension-for-pytorch/xpu/latest/tutorials/examples.html#training
# I removed saving a checkpoint, refactored and changed what is printed
# I increased the batch size so that there would be fewer steps (each being potentially slower)
# Then there are various profiler options integrated based on which parts you comment/uncomment.

# if you want to try line_profiler, include this import and the @profile decorator around train_step
#from line_profiler import profile

# if you want to try cProfile on just a portion of the code, include this import and add cProfile.Profile() below
#import cProfile

import torch
import torchvision

############# code changes ###############
import intel_extension_for_pytorch as ipex

############# code changes ###############

def train_model():
    LR = 0.001
    DOWNLOAD = True
    DATA = "datasets/cifar10/"

    transform = torchvision.transforms.Compose(
        [
            torchvision.transforms.Resize((224, 224)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
        ]
    )
    train_dataset = torchvision.datasets.CIFAR10(
        root=DATA,
        train=True,
        transform=transform,
        download=DOWNLOAD,
    )
    train_loader = torch.utils.data.DataLoader(dataset=train_dataset, batch_size=256)

    model = torchvision.models.resnet50()
    criterion = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(model.parameters(), lr=LR, momentum=0.9)
    model.train()
    ######################## code changes #######################
    model = model.to("xpu")
    criterion = criterion.to("xpu")
    model, optimizer = ipex.optimize(model, optimizer=optimizer)
    ######################## code changes #######################

    # uncomment this decorator if you want to try line_profiler
    #@profile
    def train_step(data, target, batch_idx):
        ########## code changes ##########
        data = data.to("xpu")
        target = target.to("xpu")
        ########## code changes ##########
        optimizer.zero_grad()
        output = model(data)
        loss = criterion(output, target)
        loss.backward()
        optimizer.step()
        if batch_idx % 10 == 0:
            print("batch %d: loss %.2f" % (batch_idx, loss))
   
    # this line is relevant if we use the PyTorch Profiler during the training loop
    # this schedule will skip the first 10 steps, then start 1 cycle (repeat=1). The cycle consists of waiting for 5 steps, warming up for 1 step, and actively tracing for 3 steps.
    my_schedule = torch.profiler.schedule(skip_first=10, wait=5, warmup=1, active=3, repeat=1)
    
    # this function is relevant if we use the PyTorch profiler on the training loop
    def trace_handler(p):
        output = p.key_averages().table(sort_by="self_cpu_time_total", row_limit=20)
        print(output)
        p.export_chrome_trace("trace_step" + str(p.step_num) + ".json")

    # uncomment this section for the newer PyTorch profiler and comment out the other loop below
    #with torch.profiler.profile(
    #    activities=[
    #        torch.profiler.ProfilerActivity.CPU, 
    #        torch.profiler.ProfilerActivity.XPU
    #        ],
    #    schedule=my_schedule, on_trace_ready=trace_handler, profile_memory=True, with_stack=True
    #) as prof:
    #    for batch_idx, (data, target) in enumerate(train_loader):
    #        train_step(data, target, batch_idx)
    #        prof.step()
   

    # uncomment this section for VTune
    #max_batches = 5
    #with torch.autograd.profiler.emit_itt():
    #    for batch_idx, (data, target) in enumerate(train_loader):
    #        with torch.profiler.itt.range(f'iteration_{batch_idx}'):
    #            train_step(data, target, batch_idx)
    #        if batch_idx >= max_batches:
    #            break

    # this loop is for normal training
    for batch_idx, (data, target) in enumerate(train_loader):
        train_step(data, target, batch_idx)

        # if you want to use cProfile to profile just one step, use this section:
        #if batch_idx == 42:
        #    with cProfile.Profile() as pr:
        #        train_step(data, target, batch_idx)
        #    pr.print_stats('cumulative')
        #    pr.print_stats('tottime')
        #else:
        #    train_step(data, target, batch_idx)


    print("Execution finished")

if __name__ == "__main__":
    train_model()
    
    # uncomment this section to try legacy PyTorch profiler on entire train_model function
    #with torch.autograd.profiler_legacy.profile(
    #    use_xpu=True, profile_memory=True, record_shapes=True, with_stack=True,
    #    ) as prof:
    #    train_model()
    # then uncomment printing the table below and/or saving the Chrome trace

    # You could print a table of timings from the PyTorch profiler 
    #print(prof.key_averages().table(sort_by="self_cpu_time_total"))
    # and/or you could save a Chrome trace and load it in Chrome later.
    # Note, you could call the newer PyTorch profiler here instead. However, with the newer API tracing all of train_model, the trace is way too big to load in Chrome)
    #prof.export_chrome_trace("trace_file.json")
